# Copyright 2018 miruka
# This file is part of pixcat, licensed under LGPLv3.

"""Display images on a kitty terminal with optional resizing."""

__pkg_name__ = "pixcat"
__version__  = "0.1.3"
__status__   = "Development"
# __status__ = "Production"

__author__   = "miruka"
__email__    = "miruka@disroot.org"

__license__  = "LGPLv3"
